#include "WebFrontend.h"
#include <EEPROM.h>
#include "Settings.h"
#include "HTML.h"
#include "OTAUpdate.h"


WebFrontend::WebFrontend(int port) : m_webserver(port) {
  m_port = port;
}

void WebFrontend::SetDebugMode(boolean mode) {
  m_debug = mode;
}

ESP8266WebServer *WebFrontend::WebServer() {
  return &m_webserver;
}

void WebFrontend::Begin(StateManager *stateManager) {
  m_stateManager = stateManager;

  m_webserver.on("/", [this]() {
    String result;
    result += GetTop();
    result += GetNavigation();
    result += m_stateManager->GetHTML();
    result += GetBottom();
    m_webserver.send(200, "text/html", result);
  });

  m_webserver.on("/reset", [this]() {
    m_webserver.send(200, "text/html", GetRedirectToRoot());
    ESP.restart();
  });

  m_webserver.on("/state", [this]() {
    String result;
    result += m_stateManager->GetXML();
    m_webserver.send(200, "text/xml", result);
  });

  m_webserver.on("/ota", [this]() {
    String result;
    result += GetTop();
    result += GetNavigation();
    Settings settings;
    settings.Read();
    result += "<form method='get' action='ota_start'>";
    result += "Server:&nbsp";
    result += settings.Get("otaServer", "");
    result += "<br>Port:&nbsp";
    result += settings.Get("otaPort", "");
    result += "<br>url:&nbsp";
    result += settings.Get("otaURL", "");
    result += "<br><br> <input type='submit' Value='Update and restart' > </form>";
    result += GetBottom();
    m_webserver.send(200, "text/html", result);
  });

  m_webserver.on("/ota_start", [this]() {
    m_webserver.send(200, "text/html", OTAUpdate::Start());
  });

  m_webserver.on("/save", [this]() {
    Settings settings;
    
    for (byte i = 0; i < m_webserver.args(); i++) {
      settings.Add(m_webserver.argName(i), m_webserver.arg(i));
    }
            
    if (m_debug) {
      settings.Dump();
    }
    settings.Write();

    m_webserver.send(200, "text/html", GetRedirectToRoot("Settings saved"));
    
    ESP.restart();
  });

  m_webserver.on("/setup", [this]() {
    String result;
    result += GetTop();
    result += GetNavigation();

    Settings settings;
    settings.Read();
           
    result += "<form method='get' action='save'> <table>";
          
    // ctSSID
    result += "<tr> <td> <label>SSID: </label> </td> <td> <input name='ctSSID' size='120' maxlength='32' Value='";
    result += settings.Get("ctSSID", "");
    result += "'> </td> </tr>";
           
    // ctPASS
    result += "<tr> <td> <label>Password: </label> </td> <td> <input name='ctPASS' size='120' maxlength='63' Value='";
    result += settings.Get("ctPASS", "");
    result += "'> </td> </tr>";

    result += "<tr><td></td><td><br>If IP, mask or gateway is empty, DHCP will be used</td></tr>";

    // staticIP
    result += "<tr> <td> <label>IP-Address: </label> </td> <td> <input name='staticIP' size='120' maxlength='15' Value='";
    result += settings.Get("staticIP", ""); 
    result += "'> </td> </tr>";
            
    // staticMask
    result += "<tr> <td> <label>Netmask: </label> </td> <td> <input name='staticMask' size='120' maxlength='15' Value='";
    result += settings.Get("staticMask", "");
    result += "'> </td> </tr>";
            
    // staticGW
    result += "<tr> <td> <label>Gateway: </label> </td> <td> <input name='staticGW' size='120' maxlength='15' Value='";
    result += settings.Get("staticGW", "");
    result += "'> </td> </tr>";
            
    // HostName
    result += "<tr> <td> <label>Hostname: </label> </td> <td> <input name='HostName' size='120' maxlength='63' Value='";
    result += settings.Get("HostName", "LaCrosseGateway");
    result += "'> </td> </tr>";

    // Data ports
    result += "<tr> <td> <label>Data ports: </label> </td> <td>";
    result += "<input name='DataPort1' maxlength='5' Value='";
    result += settings.Get("DataPort1", "81");
    result += "'>&nbsp;";
    result += "<input name='DataPort2' maxlength='5' Value='";
    result += settings.Get("DataPort2", "");
    result += "'>&nbsp;";
    result += "<input name='DataPort3' maxlength='5' Value='";
    result += settings.Get("DataPort3", "");
    result += "'>&nbsp;";
    result += "</td> </tr>";

    // KVInterval
    result += "<tr><td></td><td><br>Use 'off' to disable KV-transmission</td></tr>";
    result += "<tr> <td> <label>KV-Interval: </label> </td> <td> <input name='KVInterval' size='120' maxlength='3' Value='";
    result += settings.Get("KVInterval", "10");
    result += "'> </td> </tr>";

    // KVIdentity
    result += "<tr> <td> <label>KV-Identity: </label> </td> <td> <input name='KVIdentity' size='120' maxlength='20' Value='";
    result += settings.Get("KVIdentity", String(ESP.getChipId()));
    result += "'> </td> </tr>";
            
    result += "<tr><td></td><td><br>OTA update</td></tr>";

    // OTA-Server
    result += "<tr> <td> <label>OTA-Server: </label> </td> <td> <input name='otaServer' size='120' maxlength='40' Value='";
    result += settings.Get("otaServer", "");
    result += "'> </td> </tr>";

    // OTA-Port
    result += "<tr> <td> <label>OTA-Port: </label> </td> <td> <input name='otaPort' size='120' maxlength='5' Value='";
    result += settings.Get("otaPort", "");
    result += "'> </td> </tr>";

    // OTA-url
    result += "<tr> <td> <label>OTA-url: </label> </td> <td> <input name='otaURL' size='120' maxlength='80' Value='";
    result += settings.Get("otaURL", "");
    result += "'> </td> </tr>";

    // PCA301
    result += "<tr><td></td><td><br>PCA301 (normally no need to change it)</td></tr>";
    result += "<tr> <td> <label>Plugs: </label> </td> <td> <input name='PCA301Plugs' size='120' maxlength='160' Value='";
    result += settings.Get("PCA301Plugs", "");
    result += "'> </td> </tr>";

    // Flags
    result += "<tr><td></td><td><br>Only for development</td></tr>";
    result += "<tr> <td> <label>Flags: </label> </td> <td> <input name='Flags' size='120' maxlength='80' Value='";
    result += settings.Get("Flags", "");
    result += "'> </td> </tr>";

    result += "</table> <br> <input type='submit' Value='Save and restart' > </form>";

    result += GetBottom();
    m_webserver.send(200, "text/html", result);
  });


  m_webserver.onNotFound([this](){
    m_webserver.send(404, "text/plain", "Not Found");
  });

  m_webserver.begin();
}

String WebFrontend::GetNavigation() {
  String result = "";
  result += "<a href='/'>Home</a>&nbsp;&nbsp;";
  result += "<a href='setup'>Setup</a>&nbsp;&nbsp;";
  result += "<a href='ota'>OTA-Update</a>&nbsp;&nbsp;";
  result += "<br><br>";
  
  return result;
}

String WebFrontend::GetTop() {
  String result;
  result += "<!DOCTYPE HTML>\r\n<html>";
  result += "<p>LaCrosseGateway V";
  result += m_stateManager->GetVersion();
  result += "</p>";
  return result;
}

String WebFrontend::GetBottom() {
  String result;
  result += "</html>";
  return result;
}

String WebFrontend::GetRedirectToRoot(String message) {
  String result;
  result += "<html><head><meta http-equiv='refresh' content='5; URL=/'></head><body>";
  result += message;
  result += "<br><br>Reboot, please wait a moment ...</body></html>";
  return result;
}


void WebFrontend::Handle() {
  m_webserver.handleClient();
}

